/******************************************************************************
*
* project name:    TI-TiltMaze
* file name:       tiltmaze.c
* initial date:    29/11/2000
* authors:         thomas.nussbaumer@gmx.net
*                  marcos.lopez@wol.es
*
* description:     2-ball-tilt-maze game with 41 levels
*
*
* NOTE: !! No animals or trees were harmed during the creation this game !!
*
*
* $Id: tiltmaze.c,v 1.4 2004/07/14 14:59:57 DEBROUX Lionel Exp $
*
******************************************************************************/

#include "tiltmaze.h"
#include "extgraph.h"

// the ball sprite
unsigned short ball_sprite[12] = {
0x0F00,0x3FC0,0x7FE0,0x7CE0,0xFF70,0xFF70,
0xFFF0,0xFFF0,0x7FE0,0x7FE0,0x3FC0,0x0F00
};

// the destination field sprite
unsigned short dest_sprite[12] = {
0x0F00,0x30C0,0x4020,0x50A0,0x8910,0x8610,
0x8610,0x8910,0x50A0,0x4020,0x30C0,0x0F00
};

// mask for ball and destination field
unsigned short mask[12] = {
0x000F,0x000F,0x000F,0x000F,0x000F,0x000F,
0x000F,0x000F,0x000F,0x000F,0x000F,0x000F
};

short act_level = 0; // the current level

typedef struct {   // simple structure for ball and destination coordinates
    short x;
    short y;
} position_t;


position_t balls[2] = {{0,0},{1,0}};
position_t dests[2] = {{MAZE_SIZE_X-2,MAZE_SIZE_Y-1},{MAZE_SIZE_X-1,MAZE_SIZE_Y-1}};



/*===========================================================================*/
/* get input key pressed by user                                             */
/*                                                                           */
/* replacement for ngetchx() and/or GKeyIn(), because:                       */
/*                                                                           */
/* (1) ngetchx() won't check the APD timer value                             */
/* (2) GKeyIn() destroys the display due to setting the status display       */
/*                                                                           */
/* additionally this function handles the CALCULATOR OFF combinations        */
/*===========================================================================*/
unsigned short GetUserInput(void) {
    void           *kbq = kbd_queue();
    unsigned short  key;
    short           gray = GrayCheckRunning();

    //-------------------------------------------
    // restart the APD timer now
    //-------------------------------------------
    OSTimerRestart(APD_TIMER);

    while (1) {
        //---------------------------------------
        // wait until we get a key
        // if the APD timer expires -> turn
        // calculator off
        //---------------------------------------
        while (OSdequeue(&key, kbq)) {
            if (!gray) idle();
            else pokeIO(0x600005,0b10111);
            if (OSTimerExpired(APD_TIMER)) {
                off();
                OSTimerRestart(APD_TIMER);
            }
        }

        //---------------------------------------
        // handle calculator-off key combinations
        // if the APD timer expires
        // [2nd] + [ON] & [diamond]+[ON]
        // calculator off
        //---------------------------------------

        if (key == KEY_OFF || key == KEY_OFF2) {
            off();
            OSTimerRestart(APD_TIMER);
        }
        else {
            return key & ~(unsigned short)0x800; // map out key repeat flag
        }
    }
}


/*===========================================================================*/
/* utility function to draw a string to both planes                          */
/* NOTE: there exists to new styles which are handled as flags               */
/*       A_CENTERED ... the given font will be used and the string will be   */
/*                      drawn centered on the screen                         */
/*       A_SHADOWED ... if this flag is set in the attributes a lightgray    */
/*                      shadow will be drawn with +1 pixel offset in all     */
/*                      two directions                                       */
/*===========================================================================*/
void GrayDrawStrExt_R(short x,short y,const char* s,short attr,short font) {
    short do_shadow = attr & A_SHADOWED;
    attr &= ~A_SHADOWED;

    short oldfont = FontGetSys();
    FontSetSys(font);

    if (attr & A_CENTERED) {
        x = (LCD_WIDTH-DrawStrWidth(s,font))/2;
        attr &= ~A_CENTERED;
    }

    GraySetAMSPlane(DARK_PLANE);
    DrawStr(x,y,s,attr);
    GraySetAMSPlane(LIGHT_PLANE);
    DrawStr(x,y,s,attr);

    if (do_shadow) DrawStr(x+1,y+1,s,A_NORMAL);
    FontSetSys(oldfont);
}


/*===========================================================================*/
/* draw a 12x12 pixel large sprite                                           */
/*===========================================================================*/
#define DrawSprite(data,x,y)\
    Sprite16_MASK_R((x<<4)+3+PLANEOFFSET_X,(y<<4)+3+PLANEOFFSET_Y,12,data,mask,LCD_MEM)


/*===========================================================================*/
/* animate ball moving from (x1,y1) to (x2,y2)                               */
/*===========================================================================*/
static void MoveBall(short x1,short y1,short x2,short y2) {
    short redraw1 = 0;
    short redraw2 = 0;
    
    if (x1 == dests[0].x && y1 == dests[0].y) redraw1 = 1;
    if (x1 == dests[1].x && y1 == dests[1].y) redraw2 = 1;
    
    x1 = (x1<<4)+3+PLANEOFFSET_X;
    y1 = (y1<<4)+3+PLANEOFFSET_Y;
    x2 = (x2<<4)+3+PLANEOFFSET_X;
    y2 = (y2<<4)+3+PLANEOFFSET_Y;

    Sprite16_AND_R(x1,y1,12,mask,LCD_MEM);
    // while we haven't reached the destination position ...
    /*while (x1 != x2 || y1 != y2) {
        //Sprite16_AND(x1,y1,12,mask,LCD_MEM);
        if (x1 < x2) x1++;
        if (x1 > x2) x1--;
        if (y1 < y2) y1++;
        if (y1 > y2) y1--;*/
        if (redraw1) DrawSprite(dest_sprite,dests[0].x,dests[0].y);
        if (redraw2) DrawSprite(dest_sprite,dests[1].x,dests[1].y);
        //Sprite16_OR(x1,y1,12,ball_sprite,LCD_MEM);
    //}
    Sprite16_OR_R(x2,y2,12,ball_sprite,LCD_MEM);
}

// macro to access field (x,y) of level m
#define MAZE(m,x,y)  leveldata[m][(y)*MAZE_SIZE_X+(x)]


/*===========================================================================*/
/* handle a move to the left                                                 */
/*===========================================================================*/
static inline short MoveLeft(void) {
    short moved;    
    short nr_moves = 0;
    
    do {
        moved = 0;
        if (balls[0].x < balls[1].x) {    
            if (!(MAZE(act_level,balls[0].x,balls[0].y) & MASK_WALL_LEFT)) {            
                MoveBall(balls[0].x,balls[0].y,balls[0].x-1,balls[0].y/*,balls*/);
                balls[0].x--;
                moved = 1;
                nr_moves++;
            }
            if (!(MAZE(act_level,balls[1].x,balls[1].y) & MASK_WALL_LEFT) &&
                (balls[1].x-1 != balls[0].x || balls[1].y != balls[0].y))
            {
                MoveBall(balls[1].x,balls[1].y,balls[1].x-1,balls[1].y/*,balls*/);
                balls[1].x--;
                moved = 1;
                nr_moves++;
            }
        }
        else {
            if (!(MAZE(act_level,balls[1].x,balls[1].y) & MASK_WALL_LEFT)) {            
                MoveBall(balls[1].x,balls[1].y,balls[1].x-1,balls[1].y/*,balls*/);
                balls[1].x--;
                moved = 1;
                nr_moves++;
            }
            if (!(MAZE(act_level,balls[0].x,balls[0].y) & MASK_WALL_LEFT) &&
                (balls[0].x-1 != balls[1].x || balls[0].y != balls[1].y)) 
            {
                MoveBall(balls[0].x,balls[0].y,balls[0].x-1,balls[0].y/*,balls*/);
                balls[0].x--;
                nr_moves++;
                moved = 1;
            }
        }        
    } 
    while (moved);

    if (nr_moves) return 1;
    return 0;
}    


/*===========================================================================*/
/* handle a move to the right                                                */
/*===========================================================================*/
static inline short MoveRight(void) {
    short moved;
    short nr_moves = 0;
    
    do {
        moved = 0;
        if (balls[0].x > balls[1].x) {    
            if (!(MAZE(act_level,balls[0].x,balls[0].y) & MASK_WALL_RIGHT)) {            
                MoveBall(balls[0].x,balls[0].y,balls[0].x+1,balls[0].y/*,balls*/);
                balls[0].x++;
                nr_moves++;
                moved = 1;
            }
            if (!(MAZE(act_level,balls[1].x,balls[1].y) & MASK_WALL_RIGHT) &&
                (balls[1].x+1 != balls[0].x || balls[1].y != balls[0].y))
            {
                MoveBall(balls[1].x,balls[1].y,balls[1].x+1,balls[1].y/*,balls*/);
                balls[1].x++;
                nr_moves++;
                moved = 1;
            }
        }
        else {
            if (!(MAZE(act_level,balls[1].x,balls[1].y) & MASK_WALL_RIGHT)) {            
                MoveBall(balls[1].x,balls[1].y,balls[1].x+1,balls[1].y/*,balls*/);
                balls[1].x++;
                nr_moves++;
                moved = 1;
            }
            if (!(MAZE(act_level,balls[0].x,balls[0].y) & MASK_WALL_RIGHT) &&
                (balls[0].x+1 != balls[1].x || balls[0].y != balls[1].y)) 
            {
                MoveBall(balls[0].x,balls[0].y,balls[0].x+1,balls[0].y/*,balls*/);
                balls[0].x++;
                nr_moves++;
                moved = 1;
            }
        }        
    } 
    while (moved);
  
    if (nr_moves) return 1;
    return 0;
}    


/*===========================================================================*/
/* handle an upwards move                                                    */
/*===========================================================================*/
static inline short MoveUp(void) {
    short moved;
    short nr_moves = 0;
    
    do {
        moved = 0;
        if (balls[0].y < balls[1].y) {    
            if (!(MAZE(act_level,balls[0].x,balls[0].y) & MASK_WALL_TOP)) {            
                MoveBall(balls[0].x,balls[0].y,balls[0].x,balls[0].y-1/*,balls*/);
                balls[0].y--;
                nr_moves++;
                moved = 1;
            }
            if (!(MAZE(act_level,balls[1].x,balls[1].y) & MASK_WALL_TOP) &&
                (balls[1].x != balls[0].x || balls[1].y-1 != balls[0].y))
            {
                MoveBall(balls[1].x,balls[1].y,balls[1].x,balls[1].y-1/*,balls*/);
                balls[1].y--;
                nr_moves++;
                moved = 1;
            }
        }
        else {
            if (!(MAZE(act_level,balls[1].x,balls[1].y) & MASK_WALL_TOP)) {            
                MoveBall(balls[1].x,balls[1].y,balls[1].x,balls[1].y-1/*,balls*/);
                balls[1].y--;
                nr_moves++;
                moved = 1;
            }
            if (!(MAZE(act_level,balls[0].x,balls[0].y) & MASK_WALL_TOP) &&
                (balls[0].x != balls[1].x || balls[0].y-1 != balls[1].y)) 
            {
                MoveBall(balls[0].x,balls[0].y,balls[0].x,balls[0].y-1/*,balls*/);
                balls[0].y--;
                nr_moves++;
                moved = 1;
            }
        }        
    } 
    while (moved);

    if (nr_moves) return 1;
    return 0;
}    


/*===========================================================================*/
/* handle a downwards move                                                   */
/*===========================================================================*/
static inline short MoveDown(void) {
    short moved;
    short nr_moves = 0;
    
    do {
        moved = 0;
        if (balls[0].y > balls[1].y) {    
            if (!(MAZE(act_level,balls[0].x,balls[0].y) & MASK_WALL_BOTTOM)) {            
                MoveBall(balls[0].x,balls[0].y,balls[0].x,balls[0].y+1/*,balls*/);
                balls[0].y++;
                nr_moves++;
                moved = 1;
            }
            if (!(MAZE(act_level,balls[1].x,balls[1].y) & MASK_WALL_BOTTOM) &&
                (balls[1].x != balls[0].x || balls[1].y+1 != balls[0].y))
            {
                MoveBall(balls[1].x,balls[1].y,balls[1].x,balls[1].y+1/*,balls*/);
                balls[1].y++;
                nr_moves++;
                moved = 1;
            }
        }
        else {
            if (!(MAZE(act_level,balls[1].x,balls[1].y) & MASK_WALL_BOTTOM)) {            
                MoveBall(balls[1].x,balls[1].y,balls[1].x,balls[1].y+1/*,balls*/);
                balls[1].y++;
                nr_moves++;
                moved = 1;
            }
            if (!(MAZE(act_level,balls[0].x,balls[0].y) & MASK_WALL_BOTTOM) &&
                (balls[0].x != balls[1].x || balls[0].y+1 != balls[1].y)) 
            {
                MoveBall(balls[0].x,balls[0].y,balls[0].x,balls[0].y+1/*,balls*/);
                balls[0].y++;
                nr_moves++;
                moved = 1;
            }
        }        
    } 
    while (moved);

    if (nr_moves) return 1;
    return 0;
}    


/*===========================================================================*/
/* initialize level                                                          */
/*===========================================================================*/
static void InitLevel(void) {
    short x,y;
    WIN_RECT rect={PLANEOFFSET_X,
                   PLANEOFFSET_Y,
                   (MAZE_SIZE_X<<4)+1+PLANEOFFSET_X,
                   (MAZE_SIZE_Y<<4)+1+PLANEOFFSET_Y};
    
    balls[0].x   = 0;
    balls[0].y   = 0;
    balls[1].x   = 1;
    balls[1].y   = 0;

    memset(LCD_MEM,0,LCD_SIZE);

    /*rect.x0=PLANEOFFSET_X;
    rect.y0=PLANEOFFSET_Y;
    rect.x1=(MAZE_SIZE_X<<4)+1+PLANEOFFSET_X;
    rect.y1=(MAZE_SIZE_Y<<4)+1+PLANEOFFSET_Y;*/
    
    DrawClipRect(&rect, &(SCR_RECT){{0, 0, 239, 127}}, A_NORMAL); 
    
    
    for (y=0;y<MAZE_SIZE_Y;y++) {
        for (x=0;x<MAZE_SIZE_X;x++) { 
            short offset = (leveldata[act_level][y*MAZE_SIZE_X+x])<<4;
            Sprite16_OR_R((x<<4)+1+PLANEOFFSET_X,(y<<4)+1+PLANEOFFSET_Y,16,&tiles[offset],LCD_MEM);
        }
    }

    for (x=0;x<2;x++) {
        DrawSprite(dest_sprite,dests[x].x,dests[x].y);
        DrawSprite(ball_sprite,balls[x].x,balls[x].y);
    }

    FontSetSys(F_4x6);
    DrawStr(96+PLANEOFFSET_X,10+PLANEOFFSET_Y,"  \xA9 TICT 2004",A_NORMAL);
    DrawStr(85+PLANEOFFSET_X,60+PLANEOFFSET_Y,"[+][-] Change Level",  A_NORMAL);
    DrawStr(85+PLANEOFFSET_X,68+PLANEOFFSET_Y,"[CLR]    Restart Level", A_NORMAL);
    DrawStr(85+PLANEOFFSET_X,76+PLANEOFFSET_Y,"[ESC]    Exit Game",     A_NORMAL);
    FontSetSys(F_6x8);
    DrawStr(86+PLANEOFFSET_X,PLANEOFFSET_Y,PROGRAM_NAME,A_NORMAL);
    DrawStr(87+PLANEOFFSET_X,PLANEOFFSET_Y,PROGRAM_NAME,A_NORMAL);
    printf_xy(94+PLANEOFFSET_X,25+PLANEOFFSET_Y,"Level: %d",act_level+1);
}


/*===========================================================================*/
/* checks if both balls are on the destination fields                        */
/*===========================================================================*/
static inline short CheckWin(void) {
    if ((balls[0].x == dests[0].x && balls[0].y == dests[0].y) || 
        (balls[0].x == dests[1].x && balls[0].y == dests[1].y)) 
    {
        if ((balls[1].x == dests[0].x && balls[1].y == dests[0].y) || 
            (balls[1].x == dests[1].x && balls[1].y == dests[1].y)) 
        {
            return 1;
        }        
    }
    
    return 0;
}


/*===========================================================================*/
/* ... as usual: our main routine                                            */
/*===========================================================================*/
void _main(void) {
    short input;
    short moves = 0;
    //position_t* ball=(position_t *)&balls;
    //position_t* dest=(position_t *)&dests;
    
    //--------------------------------------------------------------------
    // turn grayscale on, display logo screen and turn grayscale off again
    //--------------------------------------------------------------------
    if (!GrayOn()) return;
    Logo();
    GrayOff();

    //-------------------------------------
    // initialize level and enter main loop
    //-------------------------------------        
    InitLevel(/*ball*/);
    
    do {
        //-------------------------------------------------------
        // update number of moves display and wait for user input
        //-------------------------------------------------------
        if (moves < 100) {
            printf_xy(94+PLANEOFFSET_X,35+PLANEOFFSET_Y,"Moves: %d",moves);
            input = GetUserInput();
        }
        else {
            DrawStr(7+PLANEOFFSET_X,87+PLANEOFFSET_Y,"Too many moves !",A_REPLACE);
            GetUserInput();
            input = KEY_CLEAR;
        }
        //-----------------------------------------
        // handle cursor keys and restart level key
        //-----------------------------------------
        if (input == KEY_LEFT)       moves += MoveLeft(/*ball*/); 
        else if (input == KEY_RIGHT) moves += MoveRight(/*ball*/);
        else if (input == KEY_UP)    moves += MoveUp(/*ball*/);
        else if (input == KEY_DOWN)  moves += MoveDown(/*ball*/);
        else if (input == KEY_CLEAR) {
            InitLevel(/*ball*/);
            moves = 0;
        }

        //--------------------------------------------------
        // if user has solved the level, display message and
        // proceed to next level
        //--------------------------------------------------
        if (CheckWin(/*ball,dest*/)) {
            printf_xy(94+PLANEOFFSET_X,35+PLANEOFFSET_Y,"Moves: %d",moves);
            DrawStr(7+PLANEOFFSET_X,87+PLANEOFFSET_Y,"Excellent! Level solved!",A_REPLACE);
            GetUserInput();
            input = '+';  // schedule pseudo-input of '+' to increment level
        }

        //------------------------------------
        // handle level-up and level-down keys
        //------------------------------------
        if (input == '+') {
            act_level++;
            if (act_level >= MAZE_LEVELS) act_level = 0;
            InitLevel(/*ball*/);
            moves = 0;
        }
        else if (input == '-') {
            act_level--;
            if (act_level < 0) act_level = MAZE_LEVELS-1;
            InitLevel(/*ball*/);
            moves = 0;
        }

    }
    while (input != KEY_ESC);
}


//=============================================================================
// Revision History
//=============================================================================
//
// $Log: tiltmaze.c,v $
// Revision 1.4  2004/07/14 14:59:57  DEBROUX Lionel
// recompiled with modified TIGCC 0.95 Beta 12
//
// Revision 1.3  2004/07/12 13:14:08  DEBROUX Lionel
// (1) optimizations, now using ExtGraph
// (2) changes for TI89T/AMS3.00/TIGCC 0.95
//
// Revision 1.2  2002/11/18 08:59:44  nussbaumer
// changes for V200/AMS2.08/TIGCC 0.94
//
// Revision 1.1  2000/12/09 16:04:21  Thomas Nussbaumer
// initial version
//
//
